IMPLEMENTATION MODULE File;

FROM SYSTEM IMPORT BYTE;
IMPORT IO, FIO, FIOStd, Str;
IMPORT Lib2, MD5, SO, Config;

CONST
  hNull ::= FIOStd.inv;

TYPE
  TFileDesc = RECORD
    s: TName;
    f: UFTP.TFgmLen;
    l: UFTP.TFileLen;
    d: MD5.TDgstDesc;
    n: UFTP.TFgmIdx;   (* fgm cnt *)
    e: UFTP.TFgmLen;   (* last fgm *)
    h: FIO.File;
    p: UFTP.TFileLen;  (* pos *)
  END;

VAR
  data: TFileDesc;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  IO.WrStr( s );
  IO.WrLn();
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Message( s );
  HALT;  (* solve return code problem before removing this line *)
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Message( s );
  HALT;
END Fatal;


PROCEDURE Open( s-: ARRAY OF CHAR; f: UFTP.TFgmLen; l: UFTP.TFileLen; d-: MD5.TDgstDesc );
VAR
  b: SO.TBuffer;
  t: ARRAY [0..127] OF CHAR;
BEGIN
  IF data.h # hNull THEN Fatal( 'Output file already open.' ) END;
  IF ( f < UFTP.lenMinFgm ) OR ( f > UFTP.lenMaxFgm ) THEN Fatal( 'Invalid fragment length.' ) END;
  IF ~ Config.data.ovr & FIO.Exists( s ) THEN
    SO.Init( b, t );
    SO.WrStr( b, 'File "' );
    SO.WrStr( b, s );
    SO.WrStr( b, '" exists already.' );
    Fatal( t );
  END;
  Str.Copy( data.s, s );
  data.f := f;
  data.l := l;
  data.d := d;
  data.n := l DIV f + ORD( l MOD f > 0 );
  data.e := l MOD f;
  data.h := FIO.Create( data.s );
  data.p := 0;
END Open;

PROCEDURE Write( i: UFTP.TFgmIdx; x-: ARRAY OF BYTE; l: CARDINAL );
VAR
  n: UFTP.TFgmIdx;
  p: UFTP.TFileLen;
BEGIN
  n := data.n;
  IF i >= n THEN Fatal( 'Write: Invalid fragment index.' ) END;
  DEC( n );
  IF ( i < n ) & ( l # data.f ) OR ( i = n ) & ( l # data.e ) THEN
    Fatal( 'Write: Invalid fragment length.' );
  END;
  p := i * data.f;
  IF data.p # p THEN FIO.Seek( data.h, p ) END;
  FIO.WrBin( data.h, x, l );
  data.p := FIO.GetPos( data.h );
  IF data.p # p + l THEN Fatal( 'Write: File position inconsistency.' ) END;
END Write;

PROCEDURE Close();
CONST
  len = 4 * 1024;
VAR
  c: MD5.TCtxDesc;
  b: ARRAY [1..len] OF BYTE;
  l: LONGCARD;
  n, m: CARDINAL;
  d: MD5.TDgstDesc;
BEGIN
  IF data.l # FIO.Size( data.h ) THEN Fatal( 'Close: Size mismatch.' ) END;
  FIO.Seek( data.h, 0 );
  MD5.Init( c );
  l := data.l;
  WHILE l > 0 DO
    n := len;
    IF LONGCARD( n ) > l THEN n := CARDINAL( l ) END;
    m := FIO.RdBin( data.h, b, n );
    IF m # n THEN Fatal( 'Close: Read error.' ) END;
    MD5.Update( c, b, n );
    DEC( l, LONGCARD( n ) );
  END;
  FIO.Close( data.h );
  data.h := hNull;
  MD5.Final( c, d );
  IF Lib2.Compare( d, data.d ) # SIZE( d ) THEN Fatal( 'Close: Invalid md5.' ) END;
END Close;


BEGIN
  Lib2.Fill( data, 0 );
  data.h := hNull;
END File.
